; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "Fresh artistic license."         |
;|_______________________________________________________________________________________|
;
;  Description: ForthScript engine.
;
;  Target OS: Any
;
;  Dependencies:
;
;  Notes:
;_________________________________________________________________________________________
module "Forth script compiler"

include 'ForthLib.asm'

; NOTE: E8    is the opcode for relative call
;       B8    is the opcode for mov eax, dword const
;       BA    is the opcode for mov edx, dword const
;       FF D0 is the opcode for call eax
;       C3    is the opcode for retn
;       e9    is the opcode for jmp long offset



ftypeNumber  = 0
ftypeString  = 1
ftypeStringFree = 2      ; this string should be destroyed after use.

struct TFStackCell
  .value dd ?           ; number or handle of the string.
  .type  db ?           ; number/string
  align 4
ends


struct TFLoopStackCell
ends



struct TFWordFooter
  .name    rb 16        ; max 12 bytes name paded with 0
  .entry   dd ?         ; offset to the code of the word  [.entry]-4 points to the footer of the previous word.
ends


; values for the Status flags.

forthCompileMode =  1   ; if set the script is compiled instead of interpreted
forthNextNewName =  2   ; if set the next extracted word should be interpreted as a new name.
forthAssignMode  =  4   ; if set the next variable word must pop value from the stack, instead of push.



struct TForthContext
  .pAStack     dd  ?     ; pointer to the aritmetic stack TArray.
  .pLStack     dd  ?     ; pointer to the loop stack TArray.
  .pWords      dd  ?     ; pointer to the words memory.
  .iWords      dd  ?     ; offset to the first free byte in the words memory.
  .iLastWord   dd  ?     ; offset to the end of the last word defined.
  .sizeWords   dd  ?     ; size of the words memory.

  .Status      dd ?

  .pUserWords  dd  ?     ; pointer to the user words memory.
  .iUserWords  dd  ?     ; offset of the end of the user word memory.
  .pUser       dd  ?     ; pointer to the user provided object for example TFreshEdit or other.
ends


ForthInitWordsMemory = 1024


proc ForthCreateContext
begin
        push    edi

        stdcall GetMem, sizeof.TForthContext
        mov     edi, eax

        stdcall CreateArray, sizeof.TFStackCell
        mov     [edi+TForthContext.pAStack], eax

        stdcall CreateArray, sizeof.TFLoopStackCell
        mov     [edi+TForthContext.pLStack], eax

        mov     [edi+TForthContext.sizeWords], ForthInitWordsMemory
        stdcall GetMem, [edi+TForthContext.sizeWords]
        mov     [edi+TForthContext.pWords], eax

        mov     eax, edi
        pop     edi
        return
endp


proc ForthFreeContext, .forth_context
begin


        return
endp



ferrSyntaxError = 1
ferrUnknownWord = 2
ferrOutOfMemory = 3
ferrInvalidNumber = 4
ferrInvalidString = 5
ferrMissingBracket = 6

ferrMaxError = 6

dmsgEndWithoutBegin db  'Syntax error', 0
dmsgUnknownWord     db  'Unknown word', 0
dmsgOutOfMemory     db  'Out of memory', 0
dmsgInvalidNumber   db  'Invalid number', 0
dmsgInvalidString   db  'Invalid string', 0
dmsgMissingBracket  db  'Missing ")"', 0
dmsgUnknownError    db  'Unknown error. Probably bug.', 0



ForthErrorMessages  dd  0
                    dd dmsgEndWithoutBegin
                    dd dmsgUnknownWord
                    dd dmsgOutOfMemory
                    dd dmsgInvalidNumber
                    dd dmsgInvalidString
                    dd dmsgMissingBracket
; this should be last.
                    dd dmsgUnknownError


proc GetForthErrorMsg, .errcode
begin
        mov     eax, [.errcode]
        cmp     eax, ferrMaxError
        jbe     @f
        mov     eax, ferrMaxError+1
@@:
        mov     eax, [ForthErrorMessages+4*eax]
        return
endp



proc ForthExecScript, .pcontext, .psource
.word     rb 16
.newword  rb 16
.entry    dd ?  ; the entry point of the currently compiled word.
.ofs_error dd ?
.sign      dd ?
begin
        push    ebx ecx esi edi

        mov     ebx, [.pcontext]
        mov     esi, [.psource]

.script_loop:

        xor     eax, eax
        mov     dword [.word], eax
        mov     dword [.word+4], eax
        mov     dword [.word+8], eax
        mov     dword [.word+12], eax
        xor     ecx, ecx

        dec     esi

.start_scan:

        inc     esi
        mov     al, [esi]

; offset for the error message, if any
        mov     edx, esi
        sub     edx, [.psource]
        mov     [.ofs_error], edx

        test    al, al
        jz      .end_of_script

        cmp     al, '+'
        je      .maybe_number
        cmp     al, '-'
        je      .maybe_number

        cmp     al, '0'
        jb      .nan
        cmp     al, '9'
        jbe     .extract_number

.nan:
        cmp     al, '"'
        je      .extract_string
        cmp     al, "'"
        je      .extract_string

        cmp     al, ' '
        jbe     .start_scan

        cmp     al, ':'
        jne     @f

        test    [ebx+TForthContext.Status], forthCompileMode
        jnz     .syntax_error

        or      [ebx+TForthContext.Status], forthCompileMode or forthNextNewName
        mov     eax, [ebx+TForthContext.iWords]
        mov     [.entry], eax
        jmp     .start_scan

@@:
        cmp     al, ';'
        jne     @f

        test    [ebx+TForthContext.Status], forthCompileMode
        jz      .syntax_error

; finalize the compilation of a word.

        mov     eax, sizeof.TFWordFooter+1
        call    .allocate_bytes
        jc      .err_out_of_memory

        sub     eax, sizeof.TFWordFooter
        mov     byte [eax+edx-1], $c3   ; retn

        pushd   dword [.newword] dword [.newword+4] dword [.newword+8] dword [.newword+12]
        popd    [eax+edx+12] [eax+edx+8] [eax+edx+4] [eax+edx]

        push    [.entry]
        pop     [eax+edx+TFWordFooter.entry]

        push    [ebx+TForthContext.iWords]
        pop     [ebx+TForthContext.iLastWord]

        and    [ebx+TForthContext.Status], not forthCompileMode

        jmp     .start_scan

@@:
        cmp     al, '('
        jne     @f

; comment ()
.comment_loop:
        inc     esi
        mov     al, [esi]
        test    al, al
        jz      .err_missing_bracket

        cmp     al, ')'
        jne     .comment_loop
        inc     esi
        jmp     .script_loop

@@:
        cmp     al, '\'
        jne     .extract_word

; one line comment
.comment_loop2:
        inc     esi
        mov     al, [esi]
        cmp     al, ' '
        jae     .comment_loop2
        jmp     .script_loop

.extract_word:
        mov     [.word+ecx], al
        inc     ecx
        and     ecx, $0f
        inc     esi
        mov     al, [esi]
        cmp     al, ' '
        ja      .extract_word

;.word_ok:
        test    [ebx+TForthContext.Status], forthNextNewName
        jz      .process_word

; copy the word as a new word name to be placed at the end of the word definition.
        mov     eax, dword [.word]
        mov     ecx, dword [.word+4]
        mov     dword [.newword], eax
        mov     dword [.newword+4], ecx

        mov     eax, dword [.word+8]
        mov     ecx, dword [.word+12]
        mov     dword [.newword+8], eax
        mov     dword [.newword+12], ecx

        and     [ebx+TForthContext.Status], not forthNextNewName
        jmp     .script_loop


.process_word:
        call    .search_word    ; eax offset relative to [TForthContext.pWords]
        jc      .err_unknown_word

        test    [ebx+TForthContext.Status], forthCompileMode
        jnz     .compile_the_word

        push    esi

        mov     esi, [ebx+TForthContext.pUser]
        add     eax, [ebx+TForthContext.pWords]
        call    eax

        pop     esi
        jmp     .script_loop

.compile_the_word:
        mov     ecx, eax        ; offset of the word entry

        cmp     eax, [ebx+TForthContext.iWords]
        jb      .internal_call

;.external_call:
        add     ecx, [ebx+TForthContext.pWords]         ; absolute address

        mov     eax, 7
        call    .allocate_bytes
        jc      .err_out_of_memory

        mov     byte [edx+eax-7], $b8   ; mov eax, const
        mov     [edx+eax-6], ecx
        mov     word [edx+eax-2], $d0ff

        jmp     .script_loop


.internal_call:
        mov     eax, 5
        call    .allocate_bytes ; return new offset in eax and pointer to the memory block in edx
        jc      .err_out_of_memory

; relative call to the word entry point.
        sub     ecx, eax
        mov     byte [edx+eax-5], $e8    ; call instruction opcode.
        mov     [edx+eax-4], ecx

        jmp     .script_loop


.maybe_number:
        cmp     byte [esi+1], '0'
        jb      .extract_word
        cmp     byte [esi+1], '9'
        ja      .extract_word

.extract_number:
        xor     edx, edx
        movzx   eax, al

        mov     byte [.sign], al
        cmp     al, '-'
        je      .next
        cmp     al, '+'
        je      .next

.num_loop:
        cmp     al, ' '
        jbe     .number_ok

        cmp     al, '0'
        jb      .err_invalid_number
        cmp     al, '9'
        ja      .err_invalid_number

        sub     al, '0'
        imul    edx, 10
        add     edx, eax

.next:
        inc     esi
        mov     al, [esi]
        jmp     .num_loop

.number_ok:
        cmp     byte [.sign], '-'
        jne     @f
        neg     edx
@@:
        test    [ebx+TForthContext.Status], forthCompileMode
        jnz     .compile_number

        call    forth_push_number
        jmp     .script_loop


.compile_number: ; $b8 forth_push_number $ba number $ff $d0
        mov     ecx, edx

        mov     eax, 12
        call    .allocate_bytes
        jc      .err_out_of_memory

        mov     byte  [edx+eax-12], $b8   ; mov eax, const
        mov     dword [edx+eax-11], forth_push_number
        mov     byte  [edx+eax-7], $ba    ; mov edx, const
        mov     dword [edx+eax-6], ecx
        mov     word  [edx+eax-2], $d0ff

        jmp     .script_loop


.extract_string:
        mov     ah, al          ; end character.
        xor     ecx, ecx        ; length of the string
        lea     edi, [esi+1]

.str_loop:
        inc     esi
        mov     al, [esi]
        test    al, al
        jz      .err_invalid_string

        cmp     al, ah
        je      .string_ok

        inc     ecx
        jmp     .str_loop

.string_ok:
        mov     ecx, esi
        sub     ecx, edi

        inc     esi

        test    [ebx+TForthContext.Status], forthCompileMode
        jnz     .compile_string

        stdcall StrNew
        mov     edx, eax

        stdcall StrCopyPart, edx, edi, 0, ecx

        push    edx
        stdcall AddArrayItem, [ebx+TForthContext.pAStack]
        mov     [ebx+TForthContext.pAStack], edx
        mov     [eax+TFStackCell.type], ftypeStringFree
        pop     [eax+TFStackCell.value]

        jmp     .script_loop

.compile_string:
        push    esi
        mov     esi, edi
        mov     edi, [ebx+TForthContext.iWords]

        lea     eax, [ecx+1+17]
        stdcall .allocate_bytes
        jc      .err_out_of_memory

        lea     eax, [edi+17]
        add     edi, [ebx+TForthContext.pWords]

        mov     byte [edi], $ba                ; mov edx, offset of string
        mov     dword [edi+1], eax             ; offset of string
        mov     byte [edi+5], $b8              ; mov eax, forth_push_string
        mov     dword [edi+6], forth_push_string
        mov     word [edi+10], $d0ff           ; call eax
        lea     eax, [ecx+1]
        mov     byte [edi+12], $e9             ; jmp  end
        mov     dword [edi+13], eax            ;

        add     edi, 17

        rep movsb
        mov     al, 0
        stosb

        pop     esi
        jmp     .script_loop


.end_of_script:
        mov     ecx, [ebx+TForthContext.pAStack]
        mov     eax, [ecx+TArray.count]
        test    eax, eax
        jz      @f

        lea     eax, [ecx+TArray.array+8*eax-8]
        mov     eax, [eax+TFStackCell.value]
        dec     [ecx+TArray.count]

@@:
        clc
        jmp     .finish

; error handlers
.err_missing_bracket:
        mov     eax, ferrMissingBracket
        jmp     .error

.err_invalid_string:
        mov     eax, ferrInvalidString
        jmp     .error

.err_invalid_number:
        mov     eax, ferrInvalidNumber
        jmp     .error

.err_out_of_memory:
        mov     eax, ferrOutOfMemory
        jmp     .error

.err_unknown_word:
        mov     eax, ferrUnknownWord
        jmp     .error

.syntax_error:
        mov     eax, ferrSyntaxError

.error:
        mov     edx, [.ofs_error]
        stc

.finish:
        pop     edi esi ecx ebx
        return


; internal subroutines


; returns:
; CF=0 - word found.
; eax - offset of word entry point in pWords memory.
;
; CF=1 - word not found.
.search_word:
        push    edx esi

; search in the compiled words
        mov     edx, [ebx+TForthContext.pWords]
        mov     esi, [ebx+TForthContext.iLastWord]
        call    .search_list
        jnc     .word_found

; if not found, search in the user provided words, if any.
        mov     edx, [ebx+TForthContext.pUserWords]
        test    edx, edx
        jz      .check_std

        mov     esi, [ebx+TForthContext.iUserWords]
        call    .search_list
        jnc     .word_found

.check_std:
        mov     edx, ForthStandardLib
        mov     esi, ForthStandardLib.size
        call    .search_list
        jnc     .word_found

        stc
        pop     esi edx
        retn

.word_found:
        clc
        pop     esi edx
        retn



; arguments:
;   esi - offset of the end of the word area.
;   edx - pointer to the begin of the word area.
; returns:
;   CF=0; eax=offset of the word entry in the word area.
;   CF=1 - not found
.search_list:
        add     esi, edx

.search_loop:
        sub     esi, sizeof.TFWordFooter
        cmp     esi, edx
        jb      .notfound

        mov     eax, dword [esi+TFWordFooter.name]
        cmp     eax, dword [.word]
        jne     .prev_word

        mov     eax, dword [esi+TFWordFooter.name+4]
        cmp     eax, dword [.word+4]
        jne     .prev_word

        mov     eax, dword [esi+TFWordFooter.name+8]
        cmp     eax, dword [.word+8]
        jne     .prev_word

        mov     eax, dword [esi+TFWordFooter.name+12]
        cmp     eax, dword [.word+12]
        je      .found

.prev_word:
        mov     esi, [esi+TFWordFooter.entry]
        add     esi, edx
        jmp     .search_loop

.found:
        mov     eax, [esi+TFWordFooter.entry]
        add     eax, edx
        sub     eax, [ebx+TForthContext.pWords]         ; relatively to TForthContext.pWords
        clc
        retn

.notfound:
        stc
        retn


; argument: eax - count of bytes to allocate.
; returns:
;   eax - offset in pWords of the new end ( iWords )
;   edx - pointer of pWords memory.

.allocate_bytes:
        push    ecx

        mov     edx, [ebx+TForthContext.iWords]
        add     edx, eax
        cmp     edx, [ebx+TForthContext.sizeWords]
        jb      .sizeok

        mov     ecx, [ebx+TForthContext.sizeWords]
        call    [ResizeIt]
        mov     [ebx+TForthContext.sizeWords], ecx

        stdcall ResizeMem, [ebx+TForthContext.pWords], ecx
        jc      .end_alloc
        mov     [ebx+TForthContext.pWords], eax

.sizeok:
        mov     [ebx+TForthContext.iWords], edx
        mov     eax, edx
        mov     edx, [ebx+TForthContext.pWords]
        clc

.end_alloc:
        pop     ecx
        retn

endp

; arguments:
;    ebx - TForthContext
;    edx - number to be pushed
; returns
;
forth_push_number:
        push    edx
        stdcall AddArrayItem, [ebx+TForthContext.pAStack]
        mov     [ebx+TForthContext.pAStack], edx
        mov     [eax+TFStackCell.type], ftypeNumber
        pop     [eax+TFStackCell.value]
        retn


; arguments:
;    ebx - TForthContext
; returns
;    eax - value from the top of the stack
proc forth_pop_number
begin
        mov     edx, [ebx+TForthContext.pAStack]
        mov     eax, [edx+TArray.count]
        dec     eax
        js      .empty_stack

        push    eax

        imul    eax, [edx+TArray.itemsize]
        mov     eax, [edx+TArray.array+edx+TFStackCell.value]

        stdcall DeleteArrayItem, edx ; from the stack
        clc
        return

.empty_stack:
        xor     eax, eax
        stc
        return
endp



; arguments:
;    ebx - TForthContext
;    edx - offset to the string related to pWords memory.
; returns
;
proc forth_push_string
begin
        add     edx, [ebx+TForthContext.pWords]
        push    edx
        stdcall AddArrayItem, [ebx+TForthContext.pAStack]
        mov     [ebx+TForthContext.pAStack], edx
        mov     [eax+TFStackCell.type], ftypeString
        pop     [eax+TFStackCell.value]
        return
endp






endmodule





