include '%lib%/compiler/executable.inc'

include '%lib%/macros/allmacros.inc'
include '%lib%/equates/allequates.inc'

GlobalAPI equ ascii
_BinaryType console

_CodeSection

include '%lib%/system/process.asm'
include '%lib%/simpledebug/debug.asm'
include '%lib%/system/memory.asm'
include '%lib%/system/files.asm'
include '%lib%/data/arrays.asm'
include '%lib%/data/strlib.asm'

; very simple CPU clock cycles counter:

macro START
{
        align  4
        push   ecx
        push   ebx
        cpuid  ;cpuid has different execution times the first two times
        xor    eax, eax ; Flush the pipeline
        fwait
        cpuid
        rdtsc  ;Get RDTSC counter in edx:eax
        push   edx
        push   eax
        cpuid
}

macro STOP ; CPU cycles in edx:eax
{
        fwait
        cpuid
        rdtsc
        pop    ebx
        pop    ecx
        clc
        sub    eax,ebx
        sub    eax,144h ; Cut measurement overhead 
        sbb    edx,ecx
        pop    ebx
        pop    ecx
}

start:

        InitializeAll

iglobal
        msg1    db 0ah,'Profiling FreshLib: data module',0ah,0ah,0
        sizeof.msg1 = $-msg1
        msg2    db 'This string is going to be used and abused:',0ah
                db '[español, français] {áèïôü} (1234567890)',0ah
                db '< @ # ¡ ! $ % & * - + " / | \ ', "' = ¿ ? >",0ah,0
        sizeof.msg2 = $-msg2
        msg3    db 'used ',0
        testfile db '../../UnicodeSample.txt',0
endg

uglobal
        array1  dd ?
        file1	dd ?
        string1 dd ?
        string2 dd ?
        string3 dd ?
        string4 dd ?
        string5 dd ?
        string6 dd ?
        string7 dd ?
endg

        stdcall  Output,msg1
        DebugMsg 'Cycles to create the first string:'
START
        stdcall  SetString, string1, msg1
STOP
        stdcall  OutputRegister, regEAX, 10
        DebugMsg 'Cycles to create the second string:'
START
        stdcall  SetString, string2, msg2
STOP
        stdcall  OutputRegister, regEAX, 10

        DebugMsg 'Cycles to load a file to a new string:'

START
		push     esi
		stdcall  FileOpen, testfile
		mov      [file1], eax
        stdcall  FileSeek, [file1], 0, fsFromEnd
        mov      ecx, eax
        stdcall  FileSeek, [file1], 0, fsFromBegin
		stdcall  StrNew
		mov      [string7], eax
		stdcall  StrSetCapacity, eax, ecx
		mov      esi, eax
		stdcall  FileRead, [file1], esi, ecx
		stdcall  FileClose, [file1]
		stdcall  StrFixLen, string7
STOP
        stdcall  OutputRegister, regEAX, 10
        stdcall  Output, esi
        pop      esi
        DebugMsg ''

        DebugMsg 'Cycles to extract and create a new string:'
START
        stdcall  StrExtract, [string2], 27, 15
        mov      [string4], eax
STOP
        stdcall  OutputRegister, regEAX, 10



        stdcall  StrPtr, [string4]
        stdcall  Output, eax
        DebugMsg ' <- should be: used and abused'

        stdcall  StrNew
        mov      [string6], eax
        stdcall  StrCopy, [string6], [string2]
        stdcall  SetString, string5, msg3
        stdcall  StrPos, [string6], [string5]
        mov      esi, eax
        stdcall  StrSplit, [string6], 42
        stdcall  Output, esi
        DebugMsg ' <- should be: used and abused'
        
        DebugMsg 'Cycles to measure string lenght:'
START
        stdcall  StrPtr,[string2]
        stdcall  StrLen,eax
        mov      esi, eax
STOP
        stdcall  OutputRegister, regEAX, 10
        DebugMsg 'Lenght of the string:'
        mov      eax,esi
        stdcall  OutputRegister, regEAX, 10

        DebugMsg 'Original string for the following tests:'
		stdcall  Output,msg2

        DebugMsg 'Cycles to ucase the loaded string:'
START
        stdcall  StrUCase2,[string7]
STOP
        stdcall  OutputRegister, regEAX, 10
        stdcall  StrPtr, [string7]
        stdcall  Output, eax

        DebugMsg 'Cycles to lcase the loaded string:'
START
        stdcall  StrLCase2,[string7]
STOP
        stdcall  OutputRegister, regEAX, 10
        stdcall  StrPtr, [string7]
        stdcall  Output, eax

        DebugMsg "**** testing pelaillo's case functions ****"
        stdcall  SetString,string3,msg2

        DebugMsg 'Cycles to ucase the third string:'
START
        stdcall  StrUCase2,[string3]
STOP
        stdcall  OutputRegister, regEAX, 10
        stdcall  StrPtr, [string3]
        stdcall  Output, eax

        DebugMsg 'Cycles to lcase the third string:'
START
        stdcall  StrLCase2,[string3]
STOP
        stdcall  OutputRegister, regEAX, 10
        stdcall  StrPtr, [string3]
        stdcall  Output, eax


        jmp      .exit
        stdcall  CreateArray, 20h
        jc       .error
        mov      [array1],eax
        DebugMsg 'Step2: Array Created'
        stdcall  NumToStr,eax,ntsHex
        stdcall  StrPtr, eax
        stdcall  Output, eax
        stdcall  StrNew
        mov      [string1],eax

        stdcall  AddArrayItem,[array1]
        jmp      .exit

.error:
        DebugMsg 'An error occurred'

.exit:
        FinalizeAll
        stdcall  Terminate,0

_ImportSection

include '%lib%/imports/allimports.asm'

_DataSection

IncludeAllGlobals
