
proc TButton.Create, .obj, .parent
begin
        push    eax ecx edx esi edi

        mov     ebx, [.obj]

        mov     eax, [.parent]
        test    eax, eax
        jz      .root

        mov     eax, [eax+TWindow.handle]
        jmp     .parentok

.root:
        mov     eax, [hRootWindow]

.parentok:
        cinvoke XCreateSimpleWindow, [hApplicationDisplay], eax, 0, 0, 1, 1, 0, 0, $d4d0c8
        mov     [ebx+TButton.handle], eax

        stdcall _SetWindowStruct, eax, ebx

        cinvoke XSelectInput, [hApplicationDisplay], [ebx+TButton.handle], ExposureMask or       \
                                                                           EnterWindowMask or    \
                                                                           LeaveWindowMask or    \
                                                                           ButtonPressMask or    \
                                                                           ButtonReleaseMask


        pop     edi esi edx ecx eax
        return
endp





proc TButton.SysEventHandler, .ptrEvent

begin
        push    esi edi

        mov     esi, [.ptrEvent]

        stdcall _GetWindowStruct, [esi+XEvent.window]
        mov     edi, eax

        mov     eax, [esi+XEvent.type]
        cmp     eax, Expose
        je      .expose
        cmp     eax, LeaveNotify
        je      .leave
        cmp     eax, EnterNotify
        je      .enter
        cmp     eax, ButtonPress
        je      .button_press
        cmp     eax, ButtonRelease
        je      .button_release


.finish:
        pop   edi esi
        stc
        return

.button_release:
        cmp     [esi+XButtonPressedEvent.button], 1
        jne     .finish

        cinvoke XUngrabPointer, [hApplicationDisplay], CurrentTime

        mov     [edi+TButton.state], btnPointed
        jmp     .needrepaint

.button_press:
        cmp     [esi+XButtonPressedEvent.button], 1
        jne     .finish

        cinvoke XGrabPointer, [hApplicationDisplay], [esi+XButtonEvent.window], FALSE,          \
                              ButtonReleaseMask or EnterWindowMask or LeaveWindowMask,          \
                              GrabModeAsync, GrabModeAsync, None, None, CurrentTime

        mov     [edi+TButton.state], btnPressed
        jmp     .needrepaint

.leave:
        cmp     [edi+TButton.state], btnPressed
        je      .end_mouse

        mov     [edi+TButton.state], btnNormal
        jmp     .needrepaint

.enter:
        cmp     [edi+TButton.state], btnPressed
        je      .end_mouse

        mov     [edi+TButton.state], btnPointed

.needrepaint:
locals
  .event XExposeEvent
endl
        lea     eax, [.event]
        mov     ecx, [hApplicationDisplay]
        mov     edx, [esi+XEvent.window]
        mov     [eax+XExposeEvent.type], Expose
        mov     [eax+XExposeEvent.display], ecx
        mov     [eax+XExposeEvent.window], edx

        cinvoke XSendEvent, ecx, edx, FALSE, ExposureMask, eax

.end_mouse:
        pop     edi esi
        clc
        return

.expose:
        stdcall _DrawBtnFace, [esi+XEvent.window], [edi+TButton.state]
        clc
        pop     edi esi
        return
endp







proc _DrawBtnFace, .hwin, .state
.attr XWindowAttributes
.name dd ?
.font dd ?
.direction dd ?
.ascent    dd ?
.descent   dd ?
.overall   XCharStruct

begin
        pushad

        lea     ecx, [.attr]
        cinvoke XGetWindowAttributes, [hApplicationDisplay], [.hwin], ecx
        dec     [.attr.width]
        dec     [.attr.height]

        cinvoke XCreateGC, [hApplicationDisplay], [.hwin], 0, 0
        mov     ebx, eax

; clear the background
        cinvoke XSetForeground, [hApplicationDisplay], ebx, $d4d0c8
        cinvoke XFillRectangle, [hApplicationDisplay], [.hwin], ebx, 1, 1, [.attr.width], [.attr.height]

; draw button frame
        mov     eax, $ffffff
        mov     ecx, $000000

        cmp     [.state], btnPointed
        je      .border_ok

        xchg    eax, ecx

        cmp     [.state], btnPressed
        je      .border_ok

; normal state.
        mov     eax, $e0e0e0
        mov     ecx, $808080

.border_ok:
        stdcall DrawBtnFrame, ebx, [.hwin], eax, ecx, 0, 0, [.attr.width], [.attr.height]

        cinvoke XSetForeground, [hApplicationDisplay], ebx, $000000
        cinvoke XLoadQueryFont, [hApplicationDisplay], cGUIfont
        mov     [.font], eax
        cinvoke XSetFont, [hApplicationDisplay], ebx, [eax+XFontStruct.fid]

        lea     ecx, [.name]
        cinvoke XFetchName, [hApplicationDisplay], [.hwin], ecx

        stdcall StrLength, [.name]
        mov     edi, eax
        lea     ecx, [.direction]
        lea     edx, [.ascent]
        lea     eax, [.descent]
        lea     esi, [.overall]
        cinvoke XTextExtents, [.font], [.name], edi, ecx, edx, eax, esi

        xor     eax, eax
        mov     ax, [.overall.ascent]
        add     ax, [.overall.descent]
        movsx   eax, ax
        sub     eax, [.attr.height]
        add     eax, 1
        neg     eax
        sar     eax, 1
        movsx   ecx, [.overall.ascent]
        lea     eax, [eax+ecx+1]  ; this is the baseline.
        mov     [.attr.y], eax

        mov     ax, [.overall.width]
        movsx   eax, ax
        sub     eax, [.attr.width]
        add     eax, 1
        neg     eax
        sar     eax, 1
        add     eax, 1
        cmp     eax, 1
        jge     @f
        mov     eax, 1
@@:
        mov     [.attr.x], eax

; debug only
;        cinvoke XDrawLine, [hApplicationDisplay], [.hwin], ebx, [.attr.x], 1, [.attr.x], [.attr.height]
;        cinvoke XDrawLine, [hApplicationDisplay], [.hwin], ebx, 1, [.attr.y], [.attr.width], [.attr.y]

        stdcall StrLength, [.name]
        cinvoke XDrawString, [hApplicationDisplay], [.hwin], ebx, [.attr.x], [.attr.y], [.name], eax

        cinvoke XFree, [.name]
        cinvoke XFreeFont, [hApplicationDisplay], [.font]
        cinvoke XFreeGC, [hApplicationDisplay], ebx

        popad
        return
endp


proc DrawBtnFrame, .GC, .hwin, .upcolor, .dncolor, .left, .top, .right, .bottom
begin
        cinvoke XSetForeground, [hApplicationDisplay], [.GC], [.upcolor]
        cinvoke XDrawLine, [hApplicationDisplay], [.hwin], [.GC], [.left], [.top], [.right], [.top]
        cinvoke XDrawLine, [hApplicationDisplay], [.hwin], [.GC], [.left], [.top], [.left],  [.bottom]

        cinvoke XSetForeground, [hApplicationDisplay], [.GC], [.dncolor]
        cinvoke XDrawLine, [hApplicationDisplay], [.hwin], [.GC], [.left], [.bottom], [.right], [.bottom]
        cinvoke XDrawLine, [hApplicationDisplay], [.hwin], [.GC], [.right], [.bottom], [.right],  [.top]

        return
endp






iglobal
  cGUIfont     db '*helvetica-medium-r*--12*', 0        ;'*lucida-medium-r*-sans-12*',0
endg


proc StrLength, .name
begin
        push    ecx
        mov     ecx, [.name]
        xor     eax, eax
        jecxz   .exit

        dec     eax
.loop:
        inc     eax
        cmp     byte [ecx+eax], 0
        jne     .loop
.exit:
        pop     ecx
        return
endp

