; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "Fresh artistic license."         |
;|_______________________________________________________________________________________|
;
;  Description: Asynchronous timer library.
;
;  Target OS: Any
;
;  Dependencies: memory.asm
;
;  Notes: This is OS independent part. Includes OS dependent part files as well.
;_________________________________________________________________________________________
module "Timers library"

tmfDoNothing  = 0       ; when the timer expires no action should be performed. .Expired field of the timer will be incremented.
tmfCallProc   = 1       ; [.callback] contains pointer to procedure that to be executed once per timer expiration.

tmfSyncDestroy = $10    ; If this flag is set, the timer will be destroyed on the next timer expiration. In this case, the configured event is fired and then the timer is destroyed.
                        ; The flag is checked after the event returns, so the event handler can reset this flag and thus to prevent destruction.

tmfRunning     = $80    ; If this flag is set, the timer runs. If the event handler resets this flag, the timer will fire only once and will be suspended.



struct TTimer
; internal pointers of the linked list. The field .next MUST have offset = 0
  .next dd ?            ; don't touch this

; timers parameters.
  .interval dd ?        ; in ms
  .value    dd ?        ; current value in ms. When this value becomes higher than [.interval] an event is fired and the value becomes 0.
                        ; the value is incremented by the system dependent time - probably something like 1..100ms

  .flags    dd ?        ; set of tmfXXXX values.
  .Callback dd ?        ; callback - proc Timer, .ptrTimer; TWindow or 0
  .Expired  dd ?        ; count of the timer expirations.
  .tag      dd ?        ; user value associated with the timer.
ends



uglobal
  if used __ptrFirstTimer
    __ptrFirstTimer dd ?
  end if
endg



; Creates a timer and returns pointer to TTimer structure.
; the timer is created suspended. The user can set or reset tmfRunning in [.flags] in order to start or stop the timer.
; also, the user have to enter proper values in the remaining fields.
proc TimerCreate
begin
        push    ecx

        stdcall GetMem, sizeof.TTimer
        jc      .finish

        mov     ecx, [__ptrFirstTimer]
        mov     [eax+TTimer.next], ecx
        mov     [__ptrFirstTimer], eax
        clc

.finish:
        pop     ecx
        return
endp





proc TimerDestroy, .ptrTimer
begin
        push    eax ecx

        mov     eax, [__ptrFirstTimer]
        mov     ecx, __ptrFirstTimer

        test    eax, eax
        jz      .finish                 ; if the list is empty - returns no error.

.loop:
        cmp     eax, [.ptrTimer]
        je      .found

        mov     ecx, eax
        mov     eax, [eax+TTimer.next]
        test    eax, eax
        jnz     .loop

;not found
        pop     ecx eax
        stc
        return

.found:
        push    [eax+TTimer.next]
        pop     [ecx+TTimer.next]       ; if this is the first timer, ecx == __ptrFirstTimer

        stdcall FreeMem, eax

.finish:
        pop     ecx eax
        clc
        return
endp




include '%TargetOS%/timers.asm'


endmodule



